/* 
OggCut 
	
Copyright 2002, Boris Buegling

This program is heavily based on OggDec, therefore here is the original
copyright notice:
	
"This program is distributed under the GNU General Public License, version
2. A copy of this license is included with this source.
	
Copyright 2002, Michael Smith <msmith@labyrinth.net.au>"

It also uses some code from ao/src/ao_wav.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <vorbisfile.h>

double seekme = 0;
double endme = -1;
unsigned char headbuf[44]; 

#define WRITE_U32(buf, x) *(buf)     = (unsigned char)((x)&0xff);\
                          *((buf)+1) = (unsigned char)(((x)>>8)&0xff);\
                          *((buf)+2) = (unsigned char)(((x)>>16)&0xff);\
                          *((buf)+3) = (unsigned char)(((x)>>24)&0xff);

#define WRITE_U16(buf, x) *(buf)     = (unsigned char)((x)&0xff);\
                          *((buf)+1) = (unsigned char)(((x)>>8)&0xff);

int write_prelim_header(OggVorbis_File *vf, FILE *out, ogg_int64_t knownlength) {
    unsigned int size = 0x7fffffff;
    int channels = ov_info(vf,0)->channels;
    int samplerate = ov_info(vf,0)->rate;
    int bytespersec = channels*samplerate*2;
    int align = channels*2;
    int samplesize = 16;

    if(knownlength)
        size = (unsigned int)knownlength;

    memcpy(headbuf, "RIFF", 4);
    WRITE_U32(headbuf+4, size-8);
    memcpy(headbuf+8, "WAVE", 4);
    memcpy(headbuf+12, "fmt ", 4);
    WRITE_U32(headbuf+16, 16);
    WRITE_U16(headbuf+20, 1);
    WRITE_U16(headbuf+22, channels);
    WRITE_U32(headbuf+24, samplerate);
    WRITE_U32(headbuf+28, bytespersec);
    WRITE_U16(headbuf+32, align);
    WRITE_U16(headbuf+34, samplesize);
    memcpy(headbuf+36, "data", 4);
    WRITE_U32(headbuf+40, size - 44);

    if(fwrite(headbuf, 1, 44, out) != 44) {
        fprintf(stderr, "ERROR: Failed to write wav header: %s\n", strerror(errno));
        return 1;
    }

    return 0;
}

int rewrite_header(FILE *out, unsigned int written) 
{
    unsigned int length = written;

    length += 44;

    WRITE_U32(headbuf+4, length-8);
    WRITE_U32(headbuf+40, length-44);
    if(fseek(out, 0, SEEK_SET) != 0) {
        fprintf(stderr, "ERROR: Failed to seek on seekable file: %s\n", strerror(errno));
        return 1;
    }

    if(fwrite(headbuf, 1, 44, out) != 44) {
        fprintf(stderr, "ERROR: Failed to write wav header: %s\n", strerror(errno));
        return 1;
    }
    return 0;
}

static int decode_file(char *infile, char *outfile)
{
    FILE *in, *out=NULL;
    OggVorbis_File vf;
    int bs = 0;
    char buf[8192];
    int buflen = 8192;
    unsigned int written = 0;
    unsigned int ret;
    ogg_int64_t length = 0;
    int size;
    int seekable = 0;

    if(!infile) {
        in = stdin;
    }
    else {
        in = fopen(infile, "rb");
        if(!in) {
            fprintf(stderr, "ERROR: Failed to open input file: %s\n", strerror(errno));
            return 1;
        }
    }

    if(!outfile) {
        out = stdout;
    }
    else {
        out = fopen(outfile, "wb");
        if(!in) {
            fprintf(stderr, "ERROR: Failed to open output file: %s\n", strerror(errno));
            return 1;
        }
    }

    if(ov_open(in, &vf, NULL, 0) < 0) {
        fprintf(stderr, "ERROR: Failed to open input as vorbis\n");
        fclose(in);
        fclose(out);
        return 1;
    }

    if(ov_seekable(&vf)) {
        seekable = 1;
        length = ov_pcm_total(&vf, 0);
        size = 2 * ov_info(&vf, 0)->channels;
        ov_time_seek (&vf, seekme);
	if (endme == -1)
		endme = length;
        fprintf(stderr, "Decoding \"%s\" to \"%s\"\n", infile, outfile);
    }

    if(write_prelim_header(&vf, out, length)) {
    	ov_clear(&vf);
        fclose(out);
      	return 1;
    }

    while(((ret = ov_read(&vf, buf, buflen, 0, 2, 1, &bs)) != 0) && (ov_time_tell (&vf) < endme)) {
        if(bs != 0) {
            fprintf(stderr, "Only one logical bitstream currently supported\n");
            break;
        }

        if(ret < 0) {
            fprintf(stderr, "Warning: hole in data\n");
            continue;
        }

        if(fwrite(buf, 1, ret, out) != ret) {
            fprintf(stderr, "Error writing to file: %s\n", strerror(errno));
            ov_clear(&vf);
            fclose(out);
            return 1;
        }

        written += ret;
    }

    if(seekable)
        rewrite_header(out, written); 

    ov_clear(&vf);

    fclose(out);
    return 0;
}

int main(int argc, char **argv)
{
    char *in, *out;

    if (argc > 2)
    {
	in = argv[1];
    	out = argv[2];
    }
    else
    {
    	fprintf (stderr, "Please specify at least the input and output filenames.\n");
	return 1;
    }
    
    if (argc > 3)
    	seekme = atof (argv[3]);
    if (argc > 4)
    	endme = atof (argv[4]);

    if(decode_file(in,out))
    	return 1;
    return 0;
}
